# -*- coding: utf-8 -*-
__author__ = 'Philipp D. Rohde'

import os


def open_file(path, filename):
    """
    Opens a file with the given filename at the specified path. If the path does not exist, it will be created.
    The file will be opened in write mode with UTF-8 encoding.

    :param path: path where to open the file
    :param filename: name of the file to be opened
    :return: file handler
    """
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)
    return open(path + filename, 'w', encoding='utf8')


def close_file(file):
    """
    Closes the specified file.

    :param file: the File object to close
    """
    file.close()



# -*- coding: utf-8 -*-
__author__ = 'Philipp D. Rohde'


def parse_heuristics(input_):
    """
    Parses the heuristics from the arguments passed to Trav-SHACL.

    :param input_: the heuristics argument
    :return: Python dictionary with the heuristics to be used for the evaluation
    """
    heuristics = {}
    if 'TARGET' in input_:
        heuristics['target'] = True
    else:
        heuristics['target'] = False

    if 'IN' in input_:
        heuristics['degree'] = 'in'
    elif 'OUT' in input_:
        heuristics['degree'] = 'out'
    else:
        heuristics['degree'] = False

    if 'SMALL' in input_:
        heuristics['properties'] = 'small'
    elif 'BIG' in input_:
        heuristics['properties'] = 'big'
    else:
        heuristics['properties'] = False

    return heuristics



# -*- coding: utf-8 -*-
__author__ = 'Monica Figuera'


class ValidationStats:
    """Statistics collected during the validation process."""

    def __init__(self):
        self.targets = 0
        self.valid = 0
        self.invalid = 0

        self.total_sol_mappings = 0
        self.max_sol_mappings = 0
        self.total_rules = 0
        self.max_rules = 0

        self.max_query_exec_time = 0
        self.total_query_exec_time = 0
        self.max_interleaving_time = 0
        self.total_interleaving_time = 0
        self.max_saturation_time = 0
        self.total_saturation_time = 0
        self.number_of_queries = 0
        self.totalTime = 0

        self.validation_log = ''

    def write_all_stats(self, output_file):
        """
        Writes all collected statistics to the output file.

        :param output_file: the file to be used for writing the statistics
        """
        output_file.write('all targets:\n' + str(self.targets))
        output_file.write('\nvalid targets:\n' + str(self.valid))
        output_file.write('\ninvalid targets:\n' + str(self.invalid))
        output_file.write('\nmax number of solution mappings for a query:\n' + str(self.max_sol_mappings))
        output_file.write('\ntotal number of solution mappings:\n' + str(self.total_sol_mappings))
        output_file.write('\nmax number of rules in memory:\n' + str(self.max_rules))
        output_file.write('\ntotal number of rules:\n' + str(self.total_rules))
        output_file.write('\nnumber of queries:\n' + str(self.number_of_queries))
        output_file.write('\nmax exec time for a query:\n' + str(self.max_query_exec_time))
        output_file.write('\ntotal query exec time:\n' + str(self.total_query_exec_time))
        output_file.write('\nmax interleaving (+ query exec) time for a query:\n' + str(self.max_interleaving_time))
        output_file.write('\ntotal interleaving (+ query exec) time:\n' + str(self.total_interleaving_time))
        output_file.write('\nmax (deferred) saturation time:\n' + str(self.max_saturation_time))
        output_file.write('\ntotal (deferred) saturation time:\n' + str(self.total_saturation_time))
        output_file.write('\ntotal time:\n' + str(self.totalTime) + '\n')

    def record_number_of_targets(self, valid_count, invalid_count):
        """
        Records the total number of targets for the validation.

        :param valid_count: number of valid targets
        :param invalid_count: number of invalid targets
        """
        self.valid = valid_count
        self.invalid = invalid_count
        self.targets = valid_count + invalid_count

    def record_interleaving_time(self, ms):
        """
        Records the time (in milliseconds) necessary for the interleaving process.

        :param ms: time in milliseconds passed for the interleaving step to be recorded
        """
        if ms > self.max_interleaving_time:
            self.max_interleaving_time = ms

        self.total_interleaving_time += ms

    def record_query_exec_time(self, ms):
        """
        Records the time (in milliseconds) necessary for query execution.

        :param ms: time in milliseconds passed for the query execution to be recorded
        """
        if ms > self.max_query_exec_time:
            self.max_query_exec_time = ms

        self.total_query_exec_time += ms

    def record_query(self):
        """Increases the counter for keeping track of the number of queries executed."""
        self.number_of_queries += 1

    def record_number_of_sol_mappings(self, k):
        """
        Records the number of solution mappings for the queries executed.

        :param k: the number of mappings returned by an executed query
        """
        if k > self.max_sol_mappings:
            self.max_sol_mappings = k

        self.total_sol_mappings += k

    def record_current_number_of_rules(self, k):
        """
        Records the max number of rules in memory after executing a query.

        :param k: current number of rules in memory
        """
        if k > self.max_rules:
            self.max_rules = k

    def record_total_rules(self, k):
        """
        Records the total number of rules of the validation.

        :param k: total number of grounded rules
        """
        self.total_rules = k

    def record_saturation_time(self, ms):
        """
        Records the time (in milliseconds) necessary for saturation.

        :param ms: time in milliseconds passed for the saturation step to be recorded
        """
        if ms > self.max_saturation_time:
            self.max_saturation_time = ms

        self.total_saturation_time += ms

    def record_total_time(self, ms):
        """
        Records the overall execution time (in milliseconds) of the validation.

        :param ms: time in milliseconds passed for the total validation
        """
        self.totalTime = ms

    def update_log(self, log):
        """
        Updates the validation log.

        :param log: the new log entry to be added
        """
        self.validation_log += log

    def write_validation_log(self, validation_log_output):
        """
        Writes the validation log to file.

        :param validation_log_output: file handler for the validation log
        """
        validation_log_output.write(self.validation_log)



import json
from glob import glob

import pytest
from rdflib import Graph

from TravSHACL import parse_heuristics
from TravSHACL.core.GraphTraversal import GraphTraversal
from TravSHACL.core.ShapeSchema import ShapeSchema

TEST_ENDPOINT = 'http://localhost:8899/sparql'
TEST_GRAPH = Graph().parse('./tests/data/test.ttl')


def get_all_test_cases():
    return glob('./tests/cases/**/definitions/*.json', recursive=True)


@pytest.mark.parametrize('file', get_all_test_cases())
@pytest.mark.parametrize('prio_number', ['BIG', 'SMALL'])
@pytest.mark.parametrize('prio_degree', ['IN', 'OUT'])
@pytest.mark.parametrize('prio_target', ['TARGET', ''])
@pytest.mark.parametrize('graph_traversal', ['DFS', 'BFS'])
@pytest.mark.parametrize('selective', [True, False])
@pytest.mark.parametrize('shape_format', ['JSON', 'SHACL'])
@pytest.mark.parametrize('endpoint', [TEST_ENDPOINT, TEST_GRAPH])
def test_case(file, selective, graph_traversal, prio_target, prio_degree, prio_number, shape_format, endpoint):
    if 'sparql' in file and shape_format == 'JSON':
        pytest.skip('SPARQL constraints in JSON format are not implemented.')
    if 'or_constraint' in file and shape_format == 'JSON':
        pytest.skip('OR constraints in JSON format are not implemented.')

    with open(file, 'r') as f:
        test_definition = json.load(f)

    schema_dir = test_definition['schemaDir']
    gt_valid = sorted(test_definition['groundTruth']['valid'])
    gt_invalid = sorted(test_definition['groundTruth']['invalid'])

    shape_schema = ShapeSchema(
        schema_dir=schema_dir,
        schema_format=shape_format,
        endpoint=endpoint,
        graph_traversal=GraphTraversal.BFS if graph_traversal == 'BFS' else GraphTraversal.DFS,
        heuristics=parse_heuristics(prio_target + ' ' + prio_degree + ' ' + prio_number),
        use_selective_queries=selective,
        max_split_size=256,
        output_dir=None,
        order_by_in_queries=False,
        save_outputs=False
    )
    result = shape_schema.validate()

    res_valid = []
    res_invalid = []

    for reason, value in result.items():
        for valres, instances in value.items():
            for instance in instances:
                if valres == 'valid_instances':
                    res_valid.append(instance[1])
                else:
                    res_invalid.append(instance[1])

    res_valid = sorted(res_valid)
    res_invalid = sorted(res_invalid)

    assert gt_valid == res_valid
    assert gt_invalid == res_invalid



# -*- coding: utf-8 -*-
__author__ = 'Philipp D. Rohde'

from SPARQLWrapper import SPARQLWrapper, JSON
from rdflib import Graph


class SPARQLEndpoint:
    """Implementation of a SPARQL endpoint. This implementation serves as a wrapper in order to be able to execute
       SPARQL queries over SPARQL endpoints via the Web and RDFlib graphs, i.e., in-memory knowledge graphs."""

    instance = None

    class __SPARQLEndpoint:
        """Private class to allow simulation of a Singleton."""
        def __init__(self, endpoint):
            if type(endpoint) == str:
                self.endpoint = SPARQLWrapper(endpoint)
                self.endpoint.setReturnFormat(JSON)
            else:
                self.endpoint = endpoint

        def get_endpoint_type(self):
            return type(self.endpoint)

        def run_query(self, query_string):
            if type(self.endpoint) == SPARQLWrapper:
                self.endpoint.setQuery(query_string)
                return self.endpoint.query().convert()
            else:
                # Use own serialization for the RDFLib graph query result since their serialization is slow.
                # Additionally, we only need the value in the binding; type and datatype are not checked.
                result_raw = self.endpoint.query(query_string)
                variables = result_raw.vars
                result_dict = {
                    'head': {
                        'vars': [v.toPython()[1:] for v in variables]
                    },
                    'results': {
                        'bindings': []
                    }
                }
                for result in result_raw:
                    result_dict['results']['bindings'].append(
                        {var.toPython()[1:]: {'value': result[var].toPython()} for var in variables})
                return result_dict

    def __new__(cls, endpoint):
        if type(endpoint) not in [str, Graph]:
            raise TypeError('The SPARQL endpoint needs retrieve a URL (as string) or an in-memory RDFlib graph. ' +
                            type(endpoint) + ' given instead.')
        if not SPARQLEndpoint.instance:
            SPARQLEndpoint.instance = SPARQLEndpoint.__SPARQLEndpoint(endpoint)
        return SPARQLEndpoint.instance

    def __getattr__(self, item):
        return getattr(self.instance, item)

    def __setattr__(self, key, value):
        return setattr(self.instance, key, value)



# -*- coding: utf-8 -*-
__author__ = 'Philipp D. Rohde'

from TravSHACL.constraints.Constraint import Constraint


class SPARQLConstraint(Constraint):
    """This class represents max constraints, i.e., a constraint for the maximal occurrence of a path."""

    def __init__(self, id_, is_pos, query: str = None):
        """
        Creates a new SPARQL constraint.

        :param id_: name of the constraint
        :param query: a string representing the SPARQL query to be executed for the constraint evaluation
        """
        super().__init__(id_, is_pos, None, None, None, None, None, None, None)
        self.min = -1
        self.max = -1
        self.query = query



# -*- coding: utf-8 -*-
__author__ = 'Philipp D. Rohde'

from TravSHACL.core.GraphTraversal import GraphTraversal
from TravSHACL.core.ShapeSchema import ShapeSchema
from TravSHACL.utils import parse_heuristics


def eval_shape_schema(args):
    """
    Evaluates a SHACL shape schema against a SPARQL endpoint.

    :param args: command line arguments passed to Trav-SHACL
    """
    shape_schema = ShapeSchema(
        schema_dir=args.d,
        schema_format='JSON' if args.json else 'SHACL',
        endpoint=args.endpoint,
        graph_traversal=GraphTraversal.BFS if args.graphTraversal == 'BFS' else GraphTraversal.DFS,
        heuristics=parse_heuristics(args.heuristics),
        use_selective_queries=args.selective,
        max_split_size=args.m,
        output_dir=args.outputDir,
        order_by_in_queries=args.orderby,
        save_outputs=args.outputs
    )

    report = shape_schema.validate()  # run the evaluation of the SHACL constraints over the specified endpoint
    if not args.outputs:
        print('Report:', report)
